@@ -1,7 +1,3 @@
-from Cython.Compiler.Visitor import TreeVisitor
-from Cython.Compiler.Nodes import *
-from Cython.Compiler.ExprNodes import *
-
 """
 Serializes a Cython code tree to Cython code. This is primarily useful for
 debugging and testing purposes.
@@ -10,6 +6,9 @@ The output is in a strict format, no whitespace or comments from the input
 is preserved (and it could not be as it is not present in the code tree).
 """
 
+from Cython.Compiler.Visitor import TreeVisitor
+from Cython.Compiler.ExprNodes import *
+
 class LinesResult(object):
     def __init__(self):
         self.lines = []
@@ -398,9 +397,17 @@ class CodeWriter(DeclarationWriter):
         if isinstance(posarg, AsTupleNode):
             self.visit(posarg.arg)
         else:
-            self.comma_separated_list(posarg)
-        if node.keyword_args is not None or node.starstar_arg is not None:
-            raise Exception("Not implemented yet")
+            self.comma_separated_list(posarg.args)  # TupleNode.args
+        if node.keyword_args:
+            if isinstance(node.keyword_args, DictNode):
+                for i, (name, value) in enumerate(node.keyword_args.key_value_pairs):
+                    if i > 0:
+                        self.put(', ')
+                    self.visit(name)
+                    self.put('=')
+                    self.visit(value)
+            else:
+                raise Exception("Not implemented yet")
         self.put(u")")
 
     def visit_ExprStatNode(self, node):
@@ -510,4 +517,4 @@ class PxdWriter(DeclarationWriter):
     def visit_StatNode(self, node):
         pass
 
-    
\ No newline at end of file
+    
